from datetime import datetime as dt, timedelta
import os
import multiprocessing
import time
import pandas as pd

class ProducerConsumersEngine():
    #
    # read in the configurations
    def __init__(self, producerFunc):
        #
        # get the initial configuration
        #self.consumerStatus = multiprocessing.Queue()
        self._sentinel = 'Done'
        self._producerFunc = producerFunc
        self._jobsQueue = multiprocessing.JoinableQueue()  # start a joinable queue to pass messages
        self._handlers = {}

    def register_handler(self, func):
        self._handlers[func.__name__] = func

    def main_consumer_func(self, i, **kwargs):
        #
        # execute a list of query jobs
        #print('Start executor %d' % i)
        try:
            for jobData in iter(self._jobsQueue.get, self._sentinel):
                #
                # to enforce the schema is correct, we first copy the empty table from the schema template
                # and then append the result to this empty table
                try:
                    assert jobData.handler in self._handlers
                    for key, value in kwargs.items():
                        jobData.instructions[key] = value
                    self._handlers[jobData.handler](**jobData.instructions)
                    self._jobsQueue.task_done()

                except Exception as e:
                    print('Error {} with Jobtype: {}, Instructions: {}'.format(e, jobData.handler, jobData.instructions))
                    raise e

        except Exception as e:
            print("Error {} in self.jobsQueue.get".format(e))
            raise e

    def main_producer_func(self, numExecutors, **kwargs):
        self._producerFunc(numExecutors, self._jobsQueue, **kwargs)
        #
        # Set the sentinel for all processes.
        for i in range(numExecutors):
            self._jobsQueue.put(self._sentinel)  # indicate sentinel

    def run(self, numExecutors=0, **kwargs):
        #
        # main method
        startTime = dt.now()

        if numExecutors==0:
            numExecutors = multiprocessing.cpu_count() * 8

        producer = multiprocessing.Process(name='main_producer_func',
                                           target=self.main_producer_func,
                                           args=(numExecutors,),
                                           kwargs=kwargs)
        producer.start()
        self._jobsQueue.join()
        #
        # initate consumers
        # consumer will execute the job
        consumers = [multiprocessing.Process(name='main_consumer_func',
                                             target=self.main_consumer_func,
                                             args=(i, ),
                                             kwargs=kwargs) for i in range(numExecutors)]
        for c in consumers:
            c.start()

        while True:
            if any(c.is_alive() for c in consumers):
                time.sleep(1)
            else:
                #print('Done')
                break
    #
    # define the job class
    class PCEngineJobData():
        def __init__(self, handler, inst={}):
            self.handler = handler.__name__
            self.instructions = inst


def test():
    pass

#test()


import os
import pandas as pd
from datetime import datetime as dt
import logging

from contendo_utils import *
#from contendo_logging import *

class ContendoConfigurationManager:

    def __init__(self):
        #
        # create the cache directory if required
        ProUtils.create_path_directories('cache/a')

        self.configsheet_url_template = 'https://docs.google.com/spreadsheets/d/{docId}/export?format=csv&gid={gid}&run=1'
        __main_configuration_doc_id = '1OHAkPPMtURUWu1d8BlZqcTX1iVnPeuZ1SdFLsGrgIDY'
        self._cacheConfigs = dict()

        _domains_df = self.get_configuration_pd(__main_configuration_doc_id, 0)
        self.domainsDict = ProUtils.pandas_df_to_dict(_domains_df, 'Domain')
        _templates_df = self.get_configuration_pd(__main_configuration_doc_id, 1672371224)
        self.templateDefsDict = ProUtils.pandas_df_to_dict(_templates_df, 'TemplateName')

    @contendo_classfunction_logger
    def get_configuration_pd(self, documentId, gid, domain='Contendo.config'):
        if (documentId, gid) not in self._cacheConfigs:
            try:
                _cachefile = 'cache/{}-{}.csv'.format(domain, gid)
                if not os.path.exists(_cachefile):
                    _url = self.configsheet_url_template.format(docId = documentId, gid=gid)
                    _config_df = pd.read_csv(_url).fillna('')
                    _config_df.to_csv(_cachefile, index=False)
                else:
                    _config_df = pd.read_csv(_cachefile).fillna('')

            except Exception as e:
                logger.error('Error reading document {}'.format(_url))
                raise e

            self._cacheConfigs[(documentId, gid)] = _config_df

        return self._cacheConfigs[(documentId, gid)]


    def get_configuration_dict(self, domain, gid, key):
        assert (domain in self.domainsDict)
        _config_df = self.get_configuration_pd(self.domainsDict[domain]['DocumentId'], gid, domain)
        _config_dict = ProUtils.pandas_df_to_dict(_config_df, key)

        return _config_dict

    def get_configuration_df(self, domain, gid):
        assert (domain in self.domainsDict)
        _config_df = self.get_configuration_pd(self.domainsDict[domain]['DocumentId'], gid, domain)

        return _config_df

    def get_domain_docid(self, domain):
        return self.domainsDict[domain]['DocumentId']

@contendo_function_logger
def test():
    startTime = dt.now()
    os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = "{}/sportsight-tests.json".format(os.environ["HOME"])
    ccm = ContendoConfigurationManager()
    configDict = ccm.get_configuration_dict('Football.NFL', '1564699495', 'StatName')
    logger.info('Config dict: %s, len: %s', configDict, len(configDict.keys()))
    logger.info ('NFL document id: %s', ccm.get_domain_docid('Football.NFL'))

if __name__ == '__main__':
    contendo_logging_setup(default_level=logging.DEBUG)
    test()


import json
import os
from pathlib import Path

class ProUtils:
    def __init__(self):
        return

    @staticmethod
    def pandas_df_to_dict(in_df, key):
        ret_dict = {}
        for i, config in in_df.iterrows():
            ret_dict[config[key]] = dict(config)

        return ret_dict

    @staticmethod
    def format_string(string, instructions):
        if type(string)!=str:
            return ''
        outstring = string
        for key,value in instructions.items():
            outstring=outstring.replace('{'+key+'}', str(value))

        return outstring

    #
    # utility to add a record to a dict object
    @staticmethod
    def update_dict(inDict, updateDict, keys, suffix=''):
        for key in keys:
            if key in updateDict.keys():
                inDict['{}{}'.format(key, suffix)] = updateDict[key]
        return inDict

    #
    # utility to add a record to a dict object
    @staticmethod
    def commastring_to_liststring(commaString):
        ret = str(str.split(commaString, ',')).replace('[', '').replace(']', '')
        return ret


    @staticmethod
    def get_string_from_file(fileName):
        file = open(fileName, 'r')
        ret = file.read()
        file.close()
        return ret

    @staticmethod
    def write_string_to_file(fileName, text):
        file = open(fileName, 'w')
        file.write(text)
        file.close()
        return

    @staticmethod
    def get_dict_from_jsonfile(jsonFileName):
        jsonFile = open(jsonFileName, 'r')
        ret = json.load(jsonFile)
        jsonFile.close()
        return ret

    @staticmethod
    def save_dict_to_jsonfile(jsonFileName, dict):
        jsonFile = open(jsonFileName, 'w')
        json.dump(dict, jsonFile)
        jsonFile.close()
        return

    @staticmethod
    def create_path_directories(path):
        while not os.path.exists(Path(path).parent):
            ProUtils.create_path_directories(Path(path).parent)
            os.mkdir(Path(path).parent)

#pu = ProUtils()
#print(pu.commastring_to_liststring('a,b,c,d'))



from .contendo_logging import contendo_logging_setup
from .contendo_logging import contendo_function_logger, contendo_classfunction_logger
from .dict_to_object import DictToObject
from .ProUtils import ProUtils
from .BigqueryUtils import BigqueryUtils
from .contendo_configuration_manager import ContendoConfigurationManager
from .producer_consumers_engine import ProducerConsumersEngine



import logging

from google.cloud import bigquery
from google.cloud import storage
from gcsfs import GCSFileSystem
from google.api_core.exceptions import BadRequest

from .contendo_logging import *
from .ProUtils import ProUtils

class BigqueryUtils:
    def __init__(self, project=None):
        if project:
            self.__bigquery_client = bigquery.Client(project=project)
            self.__storage_client  = storage.Client(project=project)
        else:
            self.__bigquery_client = bigquery.Client()
            self.__storage_client  = storage.Client()

    def create_dataset(self, datasetId: str):
        #
        # Make sure the target dataset exists, or create it.
        try:
            assert datasetId in [dataset.dataset_id
                                  for dataset in list(self.__bigquery_client.list_datasets())]
        except AssertionError:  # dataset doesn't exist
            datasetReference = self.__bigquery_client.dataset(datasetId)
            dataset = bigquery.Dataset(datasetReference)
            dataset.location = 'US'
            dataset = self.__bigquery_client.create_dataset(dataset)

        return

    def upload_string_to_gcp(self, data, bucketName, targetFileName):
        bucket = self.__storage_client.get_bucket(bucketName)
        blob = bucket.blob(targetFileName)
        res = blob.upload_from_string(data)
        return 'gs://{}/{}'.format(bucketName, targetFileName)

    def read_string_from_gcp(self, bucketName, fromFileName):
        data = None
        try:
            bucket = self.__storage_client.get_bucket(bucketName)
            blob = bucket.blob(fromFileName)
            data = blob.download_as_string()
        #except storage. NotFound as e:
        #    print('Info: File not found in BigqueryUtils.read_string_from_gcp({}, {})', bucketName, fromFileName)
        except Exception as e:
            print ('Error in BigqueryUtils.read_string_from_gcp({}, {}) {}, trace {}', bucketName, fromFileName, e, type(e))

        return data

    def upload_file_to_gcp(self, bucketName, inFileName, targerFileName, timestamp=False):
        if timestamp:
            ts = str(os.path.getctime(inFileName))
            self.upload_string_to_gcp(data=ts, bucketName=bucketName, targetFileName=targerFileName+'.timestamp')
        bucket = self.__storage_client.get_bucket(bucketName)
        blob = bucket.blob(targerFileName)
        res = blob.upload_from_filename(inFileName)
        return 'gs://{}/{}'.format(bucketName, targerFileName)

    def download_from_gcp(self, bucketName, fromFileName, toFileName, checkTimestamp=False):
        try:
            bucket = self.__storage_client.get_bucket(bucketName)
            blob = bucket.blob(fromFileName)
            if os.path.exists(toFileName) and checkTimestamp:
                try:
                    ts = self.read_string_from_gcp(bucketName=bucketName, fromFileName=fromFileName+'.timestamp')
                    print (float(ts), os.path.getctime(toFileName))
                    if float(ts)<=os.path.getctime(toFileName):
                        return
                except Exception as e:
                    print('File not exists %s', fromFileName+'.timestamp')

            ProUtils.create_path_directories(toFileName)
            data = blob.download_to_filename(toFileName)
        #except storage. NotFound as e:
        #    print('Info: File not found in BigqueryUtils.read_string_from_gcp({}, {})', bucketName, fromFileName)
        except Exception as e:
            print ('Error in BigqueryUtils.read_string_from_gcp({}, {}) {}, trace {}', bucketName, fromFileName, e, type(e))
            raise e

    def create_table_from_gcp_file(self, gcpFileURI, datasetId, tableId, writeDisposition='WRITE_APPEND'):
        datasetRef = self.__bigquery_client.dataset(datasetId)
        job_config = bigquery.LoadJobConfig()
        job_config.autodetect = True
        job_config.write_disposition = writeDisposition
        job_config.source_format = bigquery.SourceFormat.NEWLINE_DELIMITED_JSON
        load_job = self.__bigquery_client.load_table_from_uri(
            gcpFileURI,
            datasetRef.table(tableId),
            job_config=job_config
        )
        return load_job.result()

    @contendo_classfunction_logger
    def create_table_from_local_file(self, localFile, datasetId, tableId, writeDisposition='WRITE_TRUNCATE', fileType=bigquery.SourceFormat.NEWLINE_DELIMITED_JSON, schema=None):
        datasetRef = self.__bigquery_client.dataset(datasetId)
        job_config = bigquery.LoadJobConfig()
        job_config.autodetect = True
        job_config.write_disposition = writeDisposition
        job_config.source_format = fileType
        #job_config.max_bad_records = 100
        if schema and False:
            tableSchema = []
            for field in schema:
                tableSchema.append(bigquery.schema.SchemaField.from_api_repr(field))
            job_config.schema = tableSchema
        fileObj = open(localFile, 'rb')
        load_job = self.__bigquery_client.load_table_from_file(
            fileObj,
            datasetRef.table(tableId),
            job_config=job_config,
        )
        fileObj.close()

        try:
            result = load_job.result()
            if load_job.errors and len(load_job.errors)>0:
                logger.error('Error occured while loading file %s, errors: \n%s', localFile, load_job.errors)
        except BadRequest as be:
            logger.error('Error loading file %s, errors: %s', localFile, load_job.errors)
            raise be

        return result

    def get_table_schema(self, datasetId, tableId):

        datasetRef = self.__bigquery_client.dataset(datasetId)
        tableRef = datasetRef.table(tableId)
        table = self.__bigquery_client.get_table(tableRef)  # API Request

        schema = []
        for schemaField in table.schema:
            schema.append(schemaField.to_api_repr())

        return schema

    def execute_query(self, query):
        query_job = self.__bigquery_client.query(query)
        result = query_job.result()
        return result

    def execute_query_to_df(self, query, fillna=''):
        query_job = self.__bigquery_client.query(query)
        ret_df = query_job.result().to_dataframe().fillna(fillna)
        return ret_df

    def execute_query_to_dict(self, query, fillna=''):
        query_job = self.__bigquery_client.query(query)
        ret_df = query_job.result().to_dataframe().fillna(fillna)
        retDict = {}
        retDict['nRows'] = ret_df.shape[0]
        retDict['Columns'] = list(ret_df.columns)
        rows = []
        for i,row in ret_df.iterrows():
            rows.append(dict(row))
        retDict['Rows'] = rows
        return retDict

    def execute_query_with_schema_and_target(self, query, targetDataset, targetTable, schemaDataset=None, schemaTable=None):

        #
        # Reset the table based on schema
        writeDisposition = 'WRITE_TRUNCATE'
        if schemaDataset is not None and schemaTable is not None:
            copyJobConfig = bigquery.CopyJobConfig()
            copyJobConfig.write_disposition = 'WRITE_TRUNCATE'
            copy_job = self.__bigquery_client.copy_table(
                self.__bigquery_client.dataset(schemaDataset).table(schemaTable),
                self.__bigquery_client.dataset(targetDataset).table(targetTable),
                job_config=copyJobConfig
            )
            writeDisposition = 'WRITE_APPEND'

        #
        # Set job configuration & execute
        job_config = bigquery.QueryJobConfig()
        job_config.destination = self.__bigquery_client.dataset(targetDataset).table(targetTable)
        job_config.write_disposition = writeDisposition
        metrics_query_job = self.__bigquery_client.query(query, job_config=job_config)
        nRows = metrics_query_job.result().total_rows
        return nRows

    #def write_dataset_to_bigquery(self, pdDataset, targetTableId):


#import os
#os.chdir('~/PycharmProjects')
#os.environ["GOOGLE_APPLICATION_CREDENTIALS"]="../sportsight-tests.json"
#bqu = BigqueryUtils()
#bqu.execute_query_with_schema_and_target('SELECT lastUpdatedOn  FROM `sportsight-tests.Baseball1.pbp_playoffs_2018` LIMIT 100', 'test_dataset', 'test_table')


from setuptools import setup

setup(name='contendo_utils',
      version='0.2.23',
      description='General python and google-cloud related utilities',
      url='https://github.com/Yahalis/contendo_core1/tree/master/contendo_utils',
      author='Yahali Sherman, contendo.ai',
      author_email='yahali@contendo.ai',
      license='MIT',
      packages=['contendo_utils'],
      install_requires=[
          'gcsfs',
          'google-cloud-bigquery',
          'google-cloud-storage',
      ],
      zip_safe=False)
